/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
/*
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

/* Minimal test application program for the dmem library */

#include "mpi.h"
#include "dmem.h"

char* a2s(long);
char* n2s(long);

static MPI_Datatype handle_type;

#define NUM_IRECVS 5

int main(int argc, char *argv[])
{
    int i, mpi_rank, mpi_size, provided;
    MPI_Aint memsize;
    char pkt0[1000], pkt1[2000], pkt2[3000], pkt4[1000], *buf;
    dmem_handle dh0, dh1, dh2;
    MPI_Comm dmem_comm;

    MPI_Init_thread(0, 0, MPI_THREAD_MULTIPLE, &provided);
    if (provided != MPI_THREAD_MULTIPLE) {
        printf("insufficient level of thread safety provided by MPI implementation\n");
        MPI_Abort(MPI_COMM_WORLD,777);
    }
    MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
    MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);
    if (mpi_size != 3) {
        printf("dmem_user requires exactly 3 processes\n");
        MPI_Finalize();
        return(0);
    }
    MPI_Comm_dup(MPI_COMM_WORLD, &dmem_comm);

    MPI_Type_vector(3, 1, 1, MPI_AINT, &handle_type);
    MPI_Type_commit(&handle_type);

    if (mpi_rank == 0) {
        int i, dummy;
        MPI_Status status;
        char part_buf[20];
        dmem_handle dh4, dh5, dh6, dh7;

        /* testing commas and rounding */
        long m;
        m = 123456789;
        printf("in: %ld, out: %s\n", m, n2s(m));

        DMEM_Init(dmem_comm, 100, NUM_IRECVS); /* manager */
        DMEM_Set_errors_return(1);

        printf("globmemtab after init:\n");
        DMEM_Dumpglobmemtab();
        MPI_Barrier(dmem_comm);

        for (i = 0; i < 1000; i++) pkt0[i] = '0';
        DMEM_Put(pkt0, 1000, dh0);
        MPI_Bcast(dh0, 1, handle_type, 0, MPI_COMM_WORLD);
        MPI_Bcast(dh1, 1, handle_type, 1, MPI_COMM_WORLD);
        MPI_Bcast(dh2, 1, handle_type, 2, MPI_COMM_WORLD);

        MPI_Recv(&dummy, 1, MPI_INT, 2, 47, MPI_COMM_WORLD, &status);
        /* printf("Calling Get_part, handle is ");
           DMEM_Print_handle(dh1); printf("\n"); */
        DMEM_Get_part(dh1, 10, 20, part_buf);
        for (i = 0; i < 20; i++) {
            if (part_buf[i] != '4') {
                printf("bad match in Get_part, %c should be '4'\n", part_buf[i]);
                break;
            }
        }
        /* sleep to allow rank 1's free to take place before copying */
        /* sleep(2); */
        /* printf("Global Memory before copy\n");
           DMEM_Dumpglobmemtab(); */
        fprintf(stderr, "        rank 0 about to copy dh0: handle: %s\n", DMEM_Sprint_handle(dh0));
        fprintf(stderr, "        data in dh0: first = %c, last = %c\n", pkt0[0], pkt0[999]);
        DMEM_Copy(dh0, dh4);
        fprintf(stderr, "        copied, dh4: handle: %s\n", DMEM_Sprint_handle(dh4));
        DMEM_Get(dh4, pkt4);
        fprintf(stderr, "        data in dh4: first = %c, last = %c\n", pkt4[0], pkt4[999]);
        DMEM_Copy(dh4, dh5);
        fprintf(stderr, "        copied again, dh5: handle: %s\n", DMEM_Sprint_handle(dh5));
        DMEM_Copy(dh5, dh6);
        fprintf(stderr, "        copied again, dh6: handle: %s\n", DMEM_Sprint_handle(dh6));
        DMEM_Copy(dh5, dh7);
        fprintf(stderr, "        copied again, dh7: handle: %s\n", DMEM_Sprint_handle(dh7));
        fprintf(stderr, "        testing access functions: dh7: srvr=%d, addr=%s, size=%s\n",
               DMEM_Get_server(dh7), a2s(DMEM_Get_address(dh7)), n2s(DMEM_Get_size(dh7)));
    }
    else if (mpi_rank == 1) {
        double ask_local_time, ask_where_time, transfer_time, memcpy_time;
        int rc;

        DMEM_Init(dmem_comm, 3000, NUM_IRECVS); 
        DMEM_Set_errors_return(1);
        MPI_Barrier(dmem_comm);
        for (i = 0; i < 2000; i++) pkt1[i] = '1';
        DMEM_Put(pkt1, 2000, dh1);

        MPI_Bcast(dh0, 1, handle_type, 0, MPI_COMM_WORLD);
        MPI_Bcast(dh1, 1, handle_type, 1, MPI_COMM_WORLD);
        MPI_Bcast(dh2, 1, handle_type, 2, MPI_COMM_WORLD);

        DMEM_Get(dh0, pkt0);
        for (i = 0; i < 1000; i++) {
            if (pkt0[i] != '0') {
                printf("bad match on pkt0[%d], %c, should be 0\n", i, pkt0[i]);
                break;
            }
        }
        DMEM_Free(dh2);
        /* test error code for invalid handles */
        rc = DMEM_Get(dh2, pkt0);
        if (rc == -1)
            printf("%d: invalid get operation: get after free\n", mpi_rank);
    }
    else if (mpi_rank == 2) {
        char update_buf[20];
        int dummy;

        DMEM_Init(dmem_comm, 4000, NUM_IRECVS); 
        DMEM_Set_errors_return(1);
        MPI_Barrier(dmem_comm);
        for (i = 0; i < 3000; i++) pkt2[i] = '2';
        DMEM_Put(pkt2, 3000, dh2);
        MPI_Bcast(dh0, 1, handle_type, 0, MPI_COMM_WORLD);
        MPI_Bcast(dh1, 1, handle_type, 1, MPI_COMM_WORLD);
        MPI_Bcast(dh2, 1, handle_type, 2, MPI_COMM_WORLD);

        for (i = 0; i < 20; i++) update_buf[i] = '4';
        DMEM_Update(dh1, 10, 20, update_buf);
        MPI_Send(&dummy, 1, MPI_INT, 0, 47, MPI_COMM_WORLD);
    }
    else {
        DMEM_Init(dmem_comm, 500, 5);
    }

    MPI_Barrier(MPI_COMM_WORLD);
    DMEM_Dumplocmemtab();
    
    DMEM_Finalize();
    MPI_Finalize();

    return 0;
}  
  
