/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
/*
 *  (C) 2001 by Argonne National Laboratory.
 */

/* test program to test whether one thread can MPI-communicate with another thread in the same process */

#include <unistd.h>
#include <pthread.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include "mpi.h"

static pthread_t master_thread_id;    /* thread that calls pthread_create */
static pthread_t slave_thread_id;     /* thread created by master */
void * slave( void *);
static int rank;

int main(int argc, char *argv[])
{
    int rc, provided;
    int tag  = 17;
    int msg  = 42;

    printf("%d: Hello, World!\n", rank);

    MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &provided);
    if (provided != MPI_THREAD_MULTIPLE) {
        printf("This program needs MPI_THREAD_MILTIPLE\n");
        MPI_Finalize();
        return(1);
    }

    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    

    master_thread_id = pthread_self();
    rc = pthread_create(&slave_thread_id, NULL, slave, NULL);
    if (rc != 0) {
        printf("master thread could not create slave thread\n");
        MPI_Finalize();
        return(1);
    }
    printf("%d: master created slave\n", rank);

    printf("%d: master sending message to slave\n", rank);
    MPI_Send(&msg, 1, MPI_INT, rank, tag, MPI_COMM_WORLD); 
    printf("%d master sent message to slave\n", rank);

    pthread_join(slave_thread_id, NULL);
    MPI_Finalize();

    return(0);
}

void *slave(void * dummy)
{
    int source, tag;
    int msg;
    MPI_Status status, status1;

    printf("%d: Hello from slave\n", rank);

    printf("%d: slave thread probing for message\n", rank);
    MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status1);
    printf("%d: slave thread probed, saw message with source %d and tag %d\n", rank, status1.MPI_SOURCE, status1.MPI_TAG);
    printf("%d: slave thread receiving message from master thread\n", rank); 
    MPI_Recv(&msg, 1, MPI_INT, status1.MPI_SOURCE, status1.MPI_TAG, MPI_COMM_WORLD, &status);
    printf("%d: slave thread received message %d from master rank %d with tag %d\n",
           rank, msg, status.MPI_SOURCE, status.MPI_TAG);

    return(0);
}
