// RMA support for SHMEM

// demo with just 2 ranks first

#include <stdio.h>

#include "mpi.h"

#define NUM_ELEMS_PER_RANK  3


int main (int argc, char *argv[])
{
    int i, j, myrank, numranks, shm_myrank, shm_numranks, dispunit;
    int *arr;
    void *baseptr;
    void *all_bases[1024];
    MPI_Win win;
    MPI_Aint array_num_bytes, sz_other;
    MPI_Comm shmcomm;

    MPI_Init(&argc,&argv);
    MPI_Comm_size(MPI_COMM_WORLD,&numranks);
    MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
    printf("myrank=%d in comm_world; commsize=%d \n",myrank,numranks);

    // NOTE we are creating a new communicator split on shared mem ***********************
    MPI_Comm_split_type(MPI_COMM_WORLD, MPI_COMM_TYPE_SHARED, 0, MPI_INFO_NULL, &shmcomm);

    MPI_Comm_size(shmcomm,&shm_numranks);
    MPI_Comm_rank(shmcomm,&shm_myrank);
    printf("myrank=%d in shmcomm; commsize=%d \n",shm_myrank,shm_numranks);

    MPI_Info win_info;
    MPI_Info_create(&win_info);
    MPI_Info_set(win_info, "alloc_shared_noncontig", "true");

    array_num_bytes = NUM_ELEMS_PER_RANK * numranks * sizeof(int);
    MPI_Win_allocate_shared(array_num_bytes, sizeof(int), MPI_INFO_NULL,
                            shmcomm, &baseptr, &win);
    MPI_Info_free(&win_info);
    printf("%d: baseptr %p\n", shm_myrank, baseptr);

    // MPI_Win_lock_all(0 , win);
    MPI_Win_lock_all(MPI_MODE_NOCHECK, win);
    // now copy data to local part of shared memory

    for (i=0; i < numranks; i++)
    {
        MPI_Win_shared_query(win, i, &sz_other, &dispunit, &(all_bases[i]) );
        printf("%d: rank %d has base %p\n", myrank, i, all_bases[i]);
    }
 
    arr = baseptr;
    for (i=0; i < NUM_ELEMS_PER_RANK; i++)
        arr[i] = shm_myrank + 1000;
 
    MPI_Win_sync(win);
    MPI_Barrier(MPI_COMM_WORLD);
    MPI_Win_sync(win);
    // now we can use shared memory (prior to unlock_all below)
 
    // if (shm_myrank == 0)
    {
        for (i=0; i < shm_numranks; i++)
        {
            arr = all_bases[i];
            for (j=0; j < NUM_ELEMS_PER_RANK; j++)
                printf("%d %d: for rank %d  arr[%d] %4d\n", myrank, shm_myrank, i, j, arr[j]);
        }
    }

    MPI_Win_unlock_all(win);
 
    MPI_Win_free(&win);
 
    MPI_Finalize();
    return 0;
}
