      use mpi
      use iso_c_binding
      implicit none
      integer   :: topsize, toprank, nodesize, noderank
      integer   :: leadersize, leaderrank
      integer   :: ierr
      integer   :: i, color, key, intdispunit, doubledispunit
      integer   :: topcomm, nodecomm, leadercomm
      integer   :: intshwin, doubleshwin
      integer, parameter :: arraysize = 100
      integer   :: intarray(arraysize)
      double precision :: doublearray(arraysize)
      integer   :: intmemsize, doublememsize
      integer(kind=MPI_ADDRESS_KIND) :: longintmemsize
      integer(kind=MPI_ADDRESS_KIND) :: longdoublememsize
      type(c_ptr) :: intbaseptr, doublebaseptr
      type(c_ptr) :: rank0_intbaseptr, rank0_doublebaseptr

      integer, pointer :: intarraybaseptr(:)
      real(8), pointer :: doublearraybaseptr(:)

      call MPI_Init(ierr)

!     create a top-level communicator
      call MPI_Comm_dup(MPI_COMM_WORLD, topcomm, ierr)
      call MPI_Comm_rank(topcomm, toprank, ierr)
      call MPI_Comm_size(topcomm, topsize, ierr)

!     Initialize the data arrays on process 0 to represent the file data  

      if (toprank == 0) then
         do i = 1, arraysize, 1
            intarray(i) = i
            doublearray(i) = i
         enddo
         write(6,*) "rank    0 set intarray(i) and doublearray(i) to i"
      endif

!     split into subcummunicators, one per shared-memory node
      key = 0
      call MPI_Comm_split_type(topcomm, MPI_COMM_TYPE_SHARED, key,        &
     &                      MPI_INFO_NULL, nodecomm, ierr)
      call MPI_Comm_rank(nodecomm, noderank, ierr)
      call MPI_Comm_size(nodecomm, nodesize, ierr)
      write(6,'(a, i4, a, i4, a, i4, a)') " rank ", noderank, " of ",     &
     &      nodesize, " in nodecomm is rank ", toprank, " in topcomm"
      
!     Create a communicator composed of all the proc0's on the nodes
      if (noderank == 0) then
         color = 0
      else
         color = 1  ! leadercomm is unused by these ranks
      endif
      call MPI_Comm_split(topcomm, color, key, leadercomm, ierr)
      call MPI_Comm_size(leadercomm, leadersize, ierr)
      call MPI_Comm_rank(leadercomm, leaderrank, ierr)
      if (color == 0) then
         write(6,'(a, i4, a, i4, a)') " rank ", leaderrank,               &
     &         " in leadercomm is rank ", toprank, " in topcomm"
      endif

!     Process 0 on each node allocates the whole shared memory window
         
      if (noderank == 0) then
         intmemsize =    arraysize * 4
         doublememsize = arraysize * 8
      else
         intmemsize =    0
         doublememsize = 0
      endif

!     Allocate two windows, one for each array

      longintmemsize = intmemsize
      longdoublememsize = doublememsize

      call MPI_Win_allocate_shared(longintmemsize, MPI_INTEGER,           &
     &        MPI_INFO_NULL, nodecomm, intbaseptr, intshwin, ierr)
      call MPI_Win_allocate_shared(longdoublememsize,                     &
     &        MPI_DOUBLE_PRECISION, MPI_INFO_NULL, nodecomm,              &
     &        doublebaseptr, doubleshwin, ierr)

!     Get address of beginning of shared memory areas acquired by proc 0

      call MPI_Win_shared_query(intshwin, 0, intmemsize,                  &
     &      intdispunit, rank0_intbaseptr, ierr)
      call MPI_Win_shared_query(doubleshwin, 0, doublememsize,            &
     &      doubledispunit, rank0_doublebaseptr, ierr)

!     Proc 0 in topcomm (also is 0 in leadercomm) copies the arrays into
!     its own shared memory and then broadcasts the arrays to the other
!     processes in leadercomm

!     First, do type conversion between C pointers and Fortran arrays

      call c_f_pointer(rank0_intbaseptr, intarraybaseptr,                &
     &                 shape=[arraysize])
      call c_f_pointer(rank0_doublebaseptr, doublearraybaseptr,          &
     &                 shape=[arraysize])

      if (toprank == 0) then
         call mem_copy(intmemsize,    intarray,    intarraybaseptr)
         call mem_copy(doublememsize, doublearray, doublearraybaseptr)
      endif 

!     Broadcast from toprank 0 to the leader on each node

      if (noderank == 0) then        ! I am a proc 0 in my nodecomm
         call MPI_Bcast(intarraybaseptr, arraysize,                       &
     &           MPI_INTEGER, 0, leadercomm, ierr)
         call MPI_Bcast(doublearraybaseptr, arraysize,                    &
     &           MPI_DOUBLE_PRECISION, 0, leadercomm, ierr)
      endif

      call MPI_Barrier(topcomm, ierr);

!     Test whether the memory is really shared

      write(6, '(a, i4, a, i4, a)') " rank ", toprank, " has ",           &
     &          intarraybaseptr(3), " at intarray(3)"
      write(6, '(a, i4, a, F5.2, a)') " rank ", toprank, " has ",         &
     &          doublearraybaseptr(4), " at doublearray(4)"

      call MPI_Finalize(ierr)
      
      stop
      end


